from copy import deepcopy
from typing import Literal, get_args, get_origin

import pytest
from pydantic import ValidationError

from fmu.datamodels.fmu_results import data, enums
from fmu.datamodels.fmu_results.fmu_results import (
    FmuResults,
    FmuResultsSchema,
    ObjectMetadata,
)
from tests.utils import _get_pydantic_models_from_annotation


def test_all_content_enums_in_anydata() -> None:
    """Test that all content enums are represented with a model in AnyData."""

    anydata_models = _get_pydantic_models_from_annotation(
        data.AnyData.model_fields["root"].annotation
    )

    content_enums_in_anydata = []
    for model in anydata_models:
        # content is used as a discriminator in AnyData and
        # should be present for all models
        assert "content" in model.model_fields
        content_annotation = model.model_fields["content"].annotation

        # check that the annotation is a Literal
        assert get_origin(content_annotation) == Literal

        # get_args will unpack the enum from the Literal
        # into a tuple, should only be one Literal value
        assert len(get_args(content_annotation)) == 1

        # the literal value should be an enum
        content_enum = get_args(content_annotation)[0]
        assert isinstance(content_enum, enums.Content)

        content_enums_in_anydata.append(content_enum)

    # finally check that all content enums are represented
    for content_enum in enums.Content:
        assert content_enum in content_enums_in_anydata

    # and that number of models in AnyData matches number of content enums
    assert len(anydata_models) == len(enums.Content)


def test_content_whitelist(fluid_contact_metadata: dict) -> None:
    """Test that validation fails when value of data.content is not in
    the whitelist."""

    # assert validation as-is
    FmuResults.model_validate(fluid_contact_metadata)

    # shall fail when content is not in whitelist
    fluid_contact_metadata["data"]["content"] = "not_valid_content"
    with pytest.raises(ValidationError):
        FmuResults.model_validate(fluid_contact_metadata)


def test_content_fluid_contact(fluid_contact_metadata: dict) -> None:
    """Test content-specific rule.

    When content == fluid_contact, require the fluid_contact field
    """

    # assert validation as-is
    FmuResults.model_validate(fluid_contact_metadata)

    # check that assumptions for the test is true
    assert fluid_contact_metadata["data"]["content"] == "fluid_contact"
    assert "fluid_contact" in fluid_contact_metadata["data"]
    assert "contact" in fluid_contact_metadata["data"]["fluid_contact"]

    # assert that data.fluid_contact is required
    _metadata = deepcopy(fluid_contact_metadata)
    del _metadata["data"]["fluid_contact"]
    with pytest.raises(ValidationError):
        FmuResults.model_validate(_metadata)

    # assert that data.fluid_contact.contact is required
    _metadata = deepcopy(fluid_contact_metadata)
    del _metadata["data"]["fluid_contact"]["contact"]
    with pytest.raises(ValidationError):
        FmuResults.model_validate(_metadata)


def test_content_field_outline(field_outline_metadata: dict) -> None:
    """Test content-specific rule.

    When content == field_outline, require the field_outline field.
    """

    # assert validation as-is
    FmuResults.model_validate(field_outline_metadata)

    # check that assumptions for the test is true
    assert field_outline_metadata["data"]["content"] == "field_outline"
    assert "field_outline" in field_outline_metadata["data"]
    assert "contact" in field_outline_metadata["data"]["field_outline"]

    # assert that data.field_outline is required
    _metadata = deepcopy(field_outline_metadata)
    del _metadata["data"]["field_outline"]
    with pytest.raises(ValidationError):
        FmuResults.model_validate(_metadata)

    # assert that data.field_region.contact is required
    _metadata = deepcopy(field_outline_metadata)
    del _metadata["data"]["field_outline"]["contact"]
    with pytest.raises(ValidationError):
        FmuResults.model_validate(_metadata)


def test_content_field_region(field_region_metadata: dict) -> None:
    """Test content-specific rule: field_region.

    When content == field_region, require the data.field_region field.
    """

    # check assumptions
    assert field_region_metadata["data"]["content"] == "field_region"
    assert "field_region" in field_region_metadata["data"]
    assert "id" in field_region_metadata["data"]["field_region"]

    # assert validation as-is
    FmuResults.model_validate(field_region_metadata)

    # assert that data.field_region is required
    _metadata = deepcopy(field_region_metadata)
    del _metadata["data"]["field_region"]
    with pytest.raises(ValidationError):
        FmuResults.model_validate(_metadata)

    # assert that data.field_region.id is required and a number
    _metadata = deepcopy(field_region_metadata)
    del _metadata["data"]["field_region"]["id"]
    with pytest.raises(ValidationError):
        FmuResults.model_validate(_metadata)

    field_region_metadata["data"]["field_region"]["id"] = "NotANumber"
    with pytest.raises(ValidationError):
        FmuResults.model_validate(field_region_metadata)


def test_content_seismic(seismic_metadata: dict) -> None:
    """Test content-specific rule: seismic.

    When content == seismic, require the data.seismic field.
    """

    # check assumptions
    assert seismic_metadata["data"]["content"] == "seismic"
    assert "seismic" in seismic_metadata["data"]

    # assert validation as-is
    FmuResults.model_validate(seismic_metadata)

    # assert that data.seismic is required
    _metadata = deepcopy(seismic_metadata)
    del _metadata["data"]["seismic"]
    with pytest.raises(ValidationError):
        FmuResults.model_validate(_metadata)


def test_data_top_and_base(seismic_metadata: dict) -> None:
    """Test require data.top and data.base.

    * Require both data.top and data.base, or none.
    """

    # assert validation as-is
    FmuResults.model_validate(seismic_metadata)

    # check that assumptions for the test is true
    assert "top" in seismic_metadata["data"]
    assert "base" in seismic_metadata["data"]

    # remove "top" - shall fail
    _metadata = deepcopy(seismic_metadata)
    del _metadata["data"]["top"]
    with pytest.raises(ValidationError):
        FmuResults.model_validate(_metadata)

    # remove "base" - shall fail
    _metadata = deepcopy(seismic_metadata)
    del _metadata["data"]["base"]
    with pytest.raises(ValidationError):
        FmuResults.model_validate(_metadata)

    # remove both - shall pass
    del seismic_metadata["data"]["top"]
    del seismic_metadata["data"]["base"]
    assert "top" not in seismic_metadata["data"]  # test assumption
    assert "base" not in seismic_metadata["data"]  # test assumption
    FmuResults.model_validate(seismic_metadata)


def test_data_time(seismic_metadata: dict) -> None:
    """Test schema logic for data.time."""

    # assert validation as-is
    FmuResults.model_validate(seismic_metadata)

    # check that assumptions for the test is true
    assert "time" in seismic_metadata["data"]

    # valid when data.time is missing
    _example = deepcopy(seismic_metadata)
    del _example["data"]["time"]
    FmuResults.model_validate(_example)

    # valid when only t0
    _example = deepcopy(seismic_metadata)
    del _example["data"]["time"]["t1"]
    assert "t0" in _example["data"]["time"]  # test assumption
    FmuResults.model_validate(_example)

    # valid without labels
    _example = deepcopy(seismic_metadata)
    del _example["data"]["time"]["t0"]["label"]
    FmuResults.model_validate(_example)

    # NOT valid when other types
    for testvalue in [
        [{"t0": "2020-10-28T14:28:02", "label": "mylabel"}],
        "2020-10-28T14:28:02",
        123,
        123.4,
    ]:
        _example = deepcopy(seismic_metadata)
        _example["data"]["time"] = testvalue
        with pytest.raises(ValidationError):
            FmuResults.model_validate(_example)


def test_data_spec(
    fluid_contact_metadata: dict,
    volumes_metadata: dict,
    field_region_metadata: dict,
) -> None:
    """Test schema logic for data.spec."""

    # assert data.spec required when class == surface
    FmuResults.model_validate(fluid_contact_metadata)
    assert fluid_contact_metadata["class"] == "surface"
    assert "spec" in fluid_contact_metadata["data"]

    del fluid_contact_metadata["data"]["spec"]
    with pytest.raises(ValidationError):
        FmuResults.model_validate(fluid_contact_metadata)

    # assert data.spec required when class == table
    FmuResults.model_validate(volumes_metadata)
    assert volumes_metadata["class"] == "table"
    assert "spec" in volumes_metadata["data"]

    del volumes_metadata["data"]["spec"]
    with pytest.raises(ValidationError):
        FmuResults.model_validate(volumes_metadata)

    # assert data.spec not required when class == polygons
    FmuResults.model_validate(field_region_metadata)
    assert field_region_metadata["class"] == "polygons"
    assert "spec" in field_region_metadata["data"]

    del field_region_metadata["data"]["spec"]
    FmuResults.model_validate(field_region_metadata)


def test_zmin_zmax_not_present_for_surfaces(fluid_contact_metadata: dict) -> None:
    """
    Test that the validation works for surface metadata without
    zmin/zmax info or with zmin/zmax = None.
    """

    # assert validation as-is
    model = FmuResults.model_validate(fluid_contact_metadata)

    # assert that bbox is 3D
    assert isinstance(model.root, ObjectMetadata)
    assert isinstance(model.root.data.root.bbox, data.BoundingBox3D)

    # assert validation works with zmin/zmax = None, bbox should be 2D
    fluid_contact_metadata["data"]["bbox"]["zmin"] = None
    fluid_contact_metadata["data"]["bbox"]["zmax"] = None
    model = FmuResults.model_validate(fluid_contact_metadata)
    assert isinstance(model.root, ObjectMetadata)
    assert isinstance(model.root.data.root.bbox, data.BoundingBox2D)

    # assert validation works without zmin/zmax, bbox should be 2D
    del fluid_contact_metadata["data"]["bbox"]["zmin"]
    del fluid_contact_metadata["data"]["bbox"]["zmax"]
    model = FmuResults.model_validate(fluid_contact_metadata)
    assert isinstance(model.root, ObjectMetadata)
    assert isinstance(model.root.data.root.bbox, data.BoundingBox2D)


def test_isodatetime_format_for_time() -> None:
    """Test that the format for timestamps is set to iso-date-time
    instead of datetime that pydantic resolves it to."""

    schema = FmuResultsSchema.dump()
    timestamp_value = schema["$defs"]["Timestamp"]["properties"]["value"]

    assert timestamp_value["format"] == "iso-date-time"



"""The conftest.py, providing magical fixtures to tests."""

import os
import sys

from pytest import Config, Parser

if "--prod" in sys.argv:
    os.environ["DEV_SCHEMA"] = ""
else:
    os.environ["DEV_SCHEMA"] = "1"


def pytest_addoption(parser: Parser) -> None:
    """Addopt '--prod' as a valid command-line argument."""

    parser.addoption(
        "--prod", action="store_true", help="Use schemas/metadata with production URLs."
    )


def pytest_configure(config: Config) -> None:
    """If '--prod' is given to pytest, use schemas/metadata with prod urls (i.e. not dev
    urls)."""

    if config.getoption("--prod"):
        os.environ["DEV_SCHEMA"] = ""
    else:
        os.environ["DEV_SCHEMA"] = "1"



"""Top-level package for fmu-datamodels."""

from .fmu_results import FmuResults, FmuResultsSchema
from .standard_results import (
    FieldOutlineResult,
    FieldOutlineSchema,
    InplaceVolumesResult,
    InplaceVolumesSchema,
    StructureDepthFaultLinesResult,
    StructureDepthFaultLinesSchema,
)

try:
    from .version import version

    __version__ = version
except ImportError:
    __version__ = "0.0.0"

__all__ = [
    "FmuResults",
    "FmuResultsSchema",
    "FieldOutlineResult",
    "FieldOutlineSchema",
    "InplaceVolumesResult",
    "InplaceVolumesSchema",
    "StructureDepthFaultLinesResult",
    "StructureDepthFaultLinesSchema",
]

schemas = [
    FmuResultsSchema,
    FieldOutlineSchema,
    InplaceVolumesSchema,
    StructureDepthFaultLinesSchema,
]



from __future__ import annotations

import datetime
import getpass
import os
import platform
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
)
from uuid import UUID

from pydantic import (
    AwareDatetime,
    BaseModel,
    Field,
    GetJsonSchemaHandler,
    RootModel,
    model_validator,
)

from fmu.datamodels.types import MD5HashStr
from fmu.datamodels.version import __version__

from . import enums

if TYPE_CHECKING:
    from pydantic_core import CoreSchema


class Asset(BaseModel):
    """The ``access.asset`` block contains information about the owner asset of
    these data."""

    name: str = Field(examples=["Drogon"])
    """A string referring to a known asset name."""


class Ssdl(BaseModel):
    """
    The ``access.ssdl`` block contains information related to SSDL.
    Note that this is kept due to legacy.
    """

    access_level: enums.Classification
    """The SSDL access level. See :class:`enums.Classification`."""

    rep_include: bool
    """Flag if this data is to be shown in REP or not."""


class Access(BaseModel):
    """
    The ``access`` block contains information related to access control for
    this data object.
    """

    asset: Asset
    """A block containing information about the owner asset of these data.
    See :class:`Asset`."""

    classification: enums.Classification | None = Field(default=None)
    """The access classification level. See :class:`enums.Classification`."""


class SsdlAccess(Access):
    """
    The ``access`` block contains information related to access control for
    this data object, with legacy SSDL settings.
    """

    ssdl: Ssdl
    """A block containing information related to SSDL. See :class:`Ssdl`."""


class File(BaseModel):
    """
    The ``file`` block contains references to this data object as a file on a disk.
    A filename in this context can be actual, or abstract. Particularly the
    ``relative_path`` is, and will most likely remain, an important identifier for
    individual file objects within an FMU case - irrespective of the existance of an
    actual file system. For this reason, the ``relative_path`` - as well as the
    ``checksum_md5`` will be generated even if a file is not saved to disk. The
    ``absolute_path`` will only be generated in the case of actually creating a file on
    disk and is not required under this schema.
    """

    absolute_path: Path | None = Field(
        default=None,
        examples=["/abs/path/share/results/maps/volantis_gp_base--depth.gri"],
    )
    """The absolute path of a file, e.g. /scratch/field/user/case/etc."""

    relative_path: Path = Field(
        examples=[
            "realization-0/iter-0/share/results/maps/volantis_gp_base--depth.gri"
        ],
    )
    """The path of a file relative to the case root."""

    runpath_relative_path: Path | None = Field(
        default=None,
        examples=["share/results/maps/volantis_gp_base--depth.gri"],
    )
    """
    The path of a file relative to the runpath root of the realization.
    For files exported with the fmu.context ``case`` the field will be None.
    """

    checksum_md5: MD5HashStr = Field(examples=["fa4d055b113ae5282796e328cde0ffa4"])
    """A valid MD5 checksum of the file."""

    size_bytes: int | None = Field(default=None)
    """Size of file object in bytes"""

    @model_validator(mode="before")
    @classmethod
    def _check_for_non_ascii_in_path(cls, values: dict) -> dict:
        if (path := values.get("absolute_path")) and not str(path).isascii():
            raise ValueError(
                f"Path has non-ascii elements which is not supported: {path}"
            )
        return values


class Aggregation(BaseModel):
    """
    The ``fmu.aggregation`` block contains information about an aggregation
    performed over an ensemble.
    """

    id: UUID = Field(examples=["15ce3b84-766f-4c93-9050-b154861f9100"])
    """The unique identifier of an aggregation."""

    operation: str
    """A string representing the type of aggregation performed."""

    realization_ids: list[int]
    """An array of realization ids included in this aggregation."""


class Workflow(BaseModel):
    """
    The ``fmu.workflow`` block refers to specific subworkflows within the large
    FMU workflow being ran. This has not been standardized, mainly due to the lack of
    programmatic access to the workflows being run in important software within FMU.

    .. note:: A key usage of ``fmu.workflow.reference`` is related to ensuring
       uniqueness of data objects.
    """

    reference: str
    """A string referring to which workflow this data object was exported by."""


class User(BaseModel):
    """The ``user`` block holds information about the user."""

    id: str = Field(examples=["peesv", "jriv"])
    """A user identity reference."""


class Case(BaseModel):
    """
    The ``fmu.case`` block contains information about the case from which this data
    object was exported.

    A case represent a set of ensembles that belong together, either by being part of
    the same run (i.e. history matching) or by being placed together by the user,
    corresponding to /scratch/<asset>/<user>/<my case name>/.

    .. note:: If an FMU data object is exported outside the case context, this block
       will not be present.
    """

    name: str = Field(examples=["MyCaseName"])
    """The name of the case."""

    user: User
    """A block holding information about the user.
    See :class:`User`."""

    uuid: UUID = Field(examples=["15ce3b84-766f-4c93-9050-b154861f9100"])
    """The unique identifier of this case. Currently made by fmu.dataio."""

    description: list[str] | None = Field(default=None)
    """A free-text description of this case."""


class Ert(BaseModel):
    """The ``fmu.ert`` block contains information about the current ert run."""

    experiment: Experiment
    """Reference to the ert experiment.
    See :class:`Experiment`."""

    simulation_mode: enums.ErtSimulationMode
    """Reference to the ert simulation mode.
    See :class:`SimulationMode`."""


class Experiment(BaseModel):
    """The ``fmu.ert.experiment`` block contains information about
    the current ert experiment run."""

    id: UUID
    """The unique identifier of this ert experiment run."""


class Ensemble(BaseModel):
    """
    The ``fmu.ensemble`` block contains information about the ensemble this data
    object belongs to.
    """

    id: int = Field(ge=0)
    """The internal identification of this ensemble, represented by an integer."""

    name: str = Field(examples=["iter-0"])
    """The name of the ensemble. This is typically reflecting the folder name on
    scratch. In ERT, custom names for ensembles are supported, e.g. "pred"."""

    uuid: UUID = Field(examples=["15ce3b84-766f-4c93-9050-b154861f9100"])
    """The unique identifier of this case. Currently made by fmu.dataio."""

    restart_from: UUID | None = Field(
        default=None,
        examples=["15ce3b84-766f-4c93-9050-b154861f9100"],
    )
    """A uuid reference to another ensemble that this ensemble was restarted
    from"""


class Model(BaseModel):
    """The ``fmu.model`` block contains information about the model used.

    .. note::
       Synonyms for "model" in this context are "template", "setup", etc. The term
       "model" is ultra-generic but was chosen before e.g. "template" as the latter
       deviates from daily communications and is, if possible, even more generic
       than "model".
    """

    description: list[str] | None = Field(default=None)
    """This is a free text description of the model setup"""

    name: str = Field(examples=["Drogon"])
    """The name of the model."""

    revision: str = Field(examples=["21.0.0.dev"])
    """The revision of the model."""


class Realization(BaseModel):
    """
    The ``fmu.realization`` block contains information about the realization this
    data object belongs to.
    """

    id: int = Field(ge=0)
    """The internal ID of the realization, represented by an integer."""

    name: str = Field(examples=["realization-0"])
    """The name of the realization. This is typically reflecting the folder name on
    scratch. We recommend to use ``fmu.realization.id`` for all usage except purely
    visual appearance."""

    uuid: UUID = Field(examples=["15ce3b84-766f-4c93-9050-b154861f9100"])
    """The universally unique identifier for this realization. It is a hash of
    ``fmu.case.uuid`` and ``fmu.ensemble.uuid`` and ``fmu.realization.id``."""

    is_reference: bool | None = Field(default=None)
    """
    Flag used to determine if this realization is tagged as a reference.

    Typically, a reference realization is one that includes prediction surfaces and
    maintains all other input parameters at their default settings. However, caution
    must be exercised when putting logic upon this field, as this is simply a selected
    realization by the user and no guarantees of what the realization represents
    can be made.

    .. note::
        Please note that users shall not set this flag in the metadata upon export;
        it is intended to be configured through interactions with the Sumo GUI.
    """


class CountryItem(BaseModel):
    """A single country in the ``smda.masterdata.country`` list of countries
    known to SMDA."""

    identifier: str = Field(examples=["Norway"])
    """Identifier known to SMDA."""

    uuid: UUID = Field(examples=["15ce3b84-766f-4c93-9050-b154861f9100"])
    """Identifier known to SMDA."""


class DiscoveryItem(BaseModel):
    """A single discovery in the ``masterdata.smda.discovery`` list of discoveries
    known to SMDA."""

    short_identifier: str = Field(examples=["SomeDiscovery"])
    """Identifier known to SMDA."""

    uuid: UUID = Field(examples=["15ce3b84-766f-4c93-9050-b154861f9100"])
    """Identifier known to SMDA."""


class FieldItem(BaseModel):
    """A single field in the ``masterdata.smda.field`` list of fields
    known to SMDA."""

    identifier: str = Field(examples=["OseFax"])
    """Identifier known to SMDA."""

    uuid: UUID = Field(examples=["15ce3b84-766f-4c93-9050-b154861f9100"])
    """Identifier known to SMDA."""


class CoordinateSystem(BaseModel):
    """The ``masterdata.smda.coordinate_system`` block contains the coordinate
    system known to SMDA."""

    identifier: str = Field(examples=["ST_WGS84_UTM37N_P32637"])
    """Identifier known to SMDA."""

    uuid: UUID = Field(examples=["15ce3b84-766f-4c93-9050-b154861f9100"])
    """Identifier known to SMDA."""


class StratigraphicColumn(BaseModel):
    """The ``masterdata.smda.stratigraphic_column`` block contains the
    stratigraphic column known to SMDA."""

    identifier: str = Field(examples=["DROGON_2020"])
    """Identifier known to SMDA."""

    uuid: UUID = Field(examples=["15ce3b84-766f-4c93-9050-b154861f9100"])
    """Identifier known to SMDA."""


class Smda(BaseModel):
    """The ``masterdata.smda`` block contains SMDA-related attributes."""

    coordinate_system: CoordinateSystem
    """Reference to coordinate system known to SMDA.
    See :class:`CoordinateSystem`."""

    country: list[CountryItem]
    """A list referring to countries known to SMDA. First item is primary.
    See :class:`CountryItem`."""

    discovery: list[DiscoveryItem]
    """A list referring to discoveries known to SMDA. First item is primary.
    See :class:`DiscoveryItem`."""

    field: list[FieldItem]
    """A list referring to fields known to SMDA. First item is primary.
    See :class:`FieldItem`."""

    stratigraphic_column: StratigraphicColumn
    """Reference to stratigraphic column known to SMDA.
    See :class:`StratigraphicColumn`."""


class Masterdata(BaseModel):
    """The ``masterdata`` block contains information related to masterdata.
    Currently, SMDA holds the masterdata.
    """

    smda: Smda
    """Block containing SMDA-related attributes. See :class:`Smda`."""


class Version(BaseModel):
    """
    A generic block that contains a string representing the version of
    something.
    """

    version: str
    """A string representing the version."""


class OperatingSystem(BaseModel):
    """
    The ``operating_system`` block contains information about the OS on which the
    ensemble was run.
    """

    hostname: str = Field(examples=["st-123.equinor.com"])
    """A string containing the network name of the machine."""

    operating_system: str = Field(examples=["Darwin-18.7.0-x86_64-i386-64bit"])
    """A string containing the name of the operating system implementation."""

    release: str = Field(examples=["18.7.0"])
    """A string containing the level of the operating system."""

    system: str = Field(examples=["GNU/Linux"])
    """A string containing the name of the operating system kernel."""

    version: str = Field(examples=["#1 SMP Tue Aug 27 21:37:59 PDT 2019"])
    """The specific release version of the system."""


# TODO: Make `fmu_dataio` and `operating_system` non-optional
#  when fmu-sumo-aggregation-service uses only fmu-dataio
class SystemInformation(BaseModel):
    """
    The ``tracklog.sysinfo`` block contains information about the system upon which
    these data were exported from.
    """

    fmu_dataio: Version | None = Field(
        alias="fmu-dataio",
        default=None,
        examples=["1.2.3"],
    )
    """The version of fmu-dataio used to export the data. See :class:`Version`."""

    komodo: Version | None = Field(
        default=None,
        examples=["2023.12.05-py38"],
    )
    """The version of Komodo in which the the ensemble was run from."""

    operating_system: OperatingSystem | None = Field(default=None)
    """The operating system from which the ensemble was started from.
    See :class:`OperatingSystem`."""


class Tracklog(RootModel):
    """The ``tracklog`` block contains a record of events recorded on these data.
    This data object describes the list of tracklog events, in addition to functionality
    for constructing a tracklog and adding new records to it.
    """

    root: list[TracklogEvent]

    def __getitem__(self, item: int) -> TracklogEvent:
        return self.root[item]

    def __iter__(
        self,
    ) -> Any:
        # Using ´Any´ as return type here as mypy is having issues
        # resolving the correct type
        return iter(self.root)

    @classmethod
    def initialize(cls) -> Tracklog:
        """Initialize the tracklog object with a list containing one
        TracklogEvent of type 'created'"""
        return cls(cls._generate_tracklog_events(enums.TrackLogEventType.created))

    def extend(self, event: enums.TrackLogEventType) -> None:
        """Extend the tracklog with a new tracklog record."""
        self.root.extend(self._generate_tracklog_events(event))

    @staticmethod
    def _generate_tracklog_events(
        event: enums.TrackLogEventType,
    ) -> list[TracklogEvent]:
        """Generate new tracklog events with the given event type"""
        komodo_release = os.environ.get(
            "KOMODO_RELEASE", os.environ.get("KOMODO_RELEASE_BACKUP", None)
        )
        komodo = (
            Version.model_construct(version=komodo_release) if komodo_release else None
        )
        return [
            TracklogEvent.model_construct(
                datetime=datetime.datetime.now(datetime.UTC),
                event=event,
                user=User.model_construct(id=getpass.getuser()),
                sysinfo=(
                    SystemInformation.model_construct(
                        fmu_dataio=Version.model_construct(version=__version__),
                        komodo=komodo,
                        operating_system=(
                            OperatingSystem.model_construct(
                                hostname=platform.node(),
                                operating_system=platform.platform(),
                                release=platform.release(),
                                system=platform.system(),
                                version=platform.version(),
                            )
                        ),
                    )
                ),
            )
        ]


class TracklogEvent(BaseModel):
    """The ``tracklog`` block contains a record of events recorded on these data.
    This data object describes a tracklog event.
    """

    datetime: AwareDatetime = Field(examples=["2020-10-28T14:28:024286Z"])
    """A datetime representation recording when the event occurred."""

    event: enums.TrackLogEventType
    """The type of event being logged. See :class:`enums.TrackLogEventType`."""

    user: User
    """The user who caused the event to happen. See :class:`User`."""

    # TODO: Make non-optional when fmu-sumo-aggregation-service uses only fmu-dataio
    sysinfo: SystemInformation | None = Field(
        default_factory=SystemInformation,
    )
    """Information about the system on which the event occurred.
    See :class:`SystemInformation`."""


class Display(BaseModel):
    """
    The ``display`` block contains information related to how this data object
    should/could be displayed. As a general rule, the consumer of data is responsible
    for figuring out how a specific data object shall be displayed. However, we use
    this block to communicate preferences from the data producers perspective.

    We also maintain this block due to legacy reasons. No data filtering logic should
    be placed on the ``display`` block.
    """

    name: str | None = Field(default=None)
    """A display-friendly version of ``data.name``."""


class Context(BaseModel):
    """
    The ``fmu.context`` block contains the FMU context in which this data object
    was produced.
    """

    stage: enums.FMUContext
    """The stage of an FMU experiment in which this data was produced.
    See :class:`enums.FMUContext`."""


class IterationContext(Context):
    """
    The ``fmu.context`` block contains the FMU context in which this data object
    was produced. Here ``stage`` is required to be ``iteration``.
    """

    stage: Literal[enums.FMUContext.iteration] = Field(
        default=enums.FMUContext.iteration
    )


class EnsembleContext(Context):
    """
    The ``fmu.context`` block contains the FMU context in which this data object
    was produced. Here ``stage`` is required to be ``ensemble``.
    """

    stage: Literal[enums.FMUContext.ensemble] = Field(default=enums.FMUContext.ensemble)


class RealizationContext(Context):
    """
    The ``fmu.context`` block contains the FMU context in which this data object
    was produced. Here ``stage`` is required to be ``realization``.
    """

    stage: Literal[enums.FMUContext.realization] = Field(
        default=enums.FMUContext.realization
    )


class FMUBase(BaseModel):
    """
    The ``fmu`` block contains all attributes specific to FMU. The idea is that the FMU
    results data model can be applied to data from *other* sources - in which the
    fmu-specific stuff may not make sense or be applicable.
    """

    case: Case
    """The ``fmu.case`` block contains information about the case from which this data
    object was exported. See :class:`Case`."""

    model: Model
    """The ``fmu.model`` block contains information about the model used.
    See :class:`Model`."""


class FMUIteration(FMUBase):
    """Deprecated and replaced by :class:`FMUEnsemble`."""

    context: IterationContext
    """The ``fmu.context`` block contains the FMU context in which this data object
    was produced. See :class:`Context`. For ``iteration`` the context is ``iteration``.
    """

    iteration: Ensemble
    """The ``fmu.iteration`` block contains information about the iteration this data
    object belongs to. See :class:`Iteration`. """


class FMUEnsemble(FMUBase):
    """
    The ``fmu`` block contains all attributes specific to FMU. The idea is that the FMU
    results data model can be applied to data from *other* sources - in which the
    fmu-specific stuff may not make sense or be applicable.
    This is a specialization of the FMU block for ``ensemble`` objects.
    """

    context: EnsembleContext
    """The ``fmu.context`` block contains the FMU context in which this data object
    was produced. See :class:`Context`. For ``ensemble`` the context is ``ensemble``.
    """

    ensemble: Ensemble
    """The ``fmu.ensemble`` block contains information about the ensemble this data
    object belongs to. See :class:`ensemble`. """


class FMURealization(FMUBase):
    """
    The ``fmu`` block contains all attributes specific to FMU. The idea is that the FMU
    results data model can be applied to data from *other* sources - in which the
    fmu-specific stuff may not make sense or be applicable.
    This is a specialization of the FMU block for ``realization`` objects.
    """

    context: RealizationContext
    """The ``fmu.context`` block contains the FMU context in which this data object
    was produced. See :class:`Context`. For ``realization`` the context is always
    ``realization``.
    """

    ensemble: Ensemble
    """The ``fmu.ensemble`` block contains information about the ensemble this data
    object belongs to. See :class:`ensemble`. """

    iteration: Ensemble | None = Field(default=None)
    """Deprecated and replaced by ``fmu.ensemble``"""

    realization: Realization
    """The ``fmu.realization`` block contains information about the realization this
    data object belongs to. See :class:`Realization`."""


class Entity(BaseModel):
    """
    The ``fmu.entity`` block identifies data objects representing the same entity
    within a case, i.e. both realizations and aggregations of a particular entity
    will share the same unique identifier.
    """

    uuid: UUID = Field(examples=["15ce3b84-766f-4c93-9050-b154861f9100"])
    """The unique identifier of an object entity within a case."""


class FMU(FMUBase):
    """
    The ``fmu`` block contains all attributes specific to FMU. The idea is that the FMU
    results data model can be applied to data from *other* sources - in which the
    fmu-specific stuff may not make sense or be applicable.
    """

    context: Context
    """The ``fmu.context`` block contains the FMU context in which this data object
    was produced. See :class:`Context`.  """

    ensemble: Ensemble | None = Field(default=None)
    """The ``fmu.ensemble`` block contains information about the ensemble this data
    object belongs to. See :class:`Ensemble`. """

    iteration: Ensemble | None = Field(default=None)
    """Deprecated and replaced by ``fmu.ensemble``"""

    workflow: Workflow | None = Field(default=None)
    """The ``fmu.workflow`` block refers to specific subworkflows within the large
    FMU workflow being ran. See :class:`Workflow`."""

    aggregation: Aggregation | None = Field(default=None)
    """The ``fmu.aggregation`` block contains information about an aggregation
    performed over an ensemble. See :class:`Aggregation`."""

    realization: Realization | None = Field(default=None)
    """The ``fmu.realization`` block contains information about the realization this
    data object belongs to. See :class:`Realization`."""

    entity: Entity | None = Field(default=None)
    """The ``fmu.entity`` block identifies data objects representing the same entity
    within a case. Note, for objects exported in context ``case`` this field will
    be empty. See :class:`Ensemble`."""

    ert: Ert | None = Field(default=None)
    """The ``fmu.ert`` block contains information about the current ert run
    See :class:`Ert`."""

    @model_validator(mode="after")
    def _set_iteration_equal_ensemble(self) -> FMU:
        """
        The 'fmu.ensemble' field has replaced the 'fmu.iteration' field. However in a
        transition period we keep both, hence we set the 'fmu.iteration' here.
        """
        self.iteration = self.ensemble
        return self

    @model_validator(mode="before")
    @classmethod
    def _dependencies_aggregation_realization(cls, values: dict) -> dict:
        aggregation, realization = values.get("aggregation"), values.get("realization")
        if aggregation and realization:
            raise ValueError(
                "Both 'aggregation' and 'realization' cannot be set "
                "at the same time. Please set only one."
            )
        return values

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        core_schema: CoreSchema,
        handler: GetJsonSchemaHandler,
    ) -> dict[str, object]:
        json_schema = super().__get_pydantic_json_schema__(core_schema, handler)
        json_schema = handler.resolve_ref_schema(json_schema)
        json_schema.update(
            {
                "dependencies": {
                    "aggregation": {"not": {"required": ["realization"]}},
                    "realization": {"not": {"required": ["aggregation"]}},
                }
            }
        )
        return json_schema



from __future__ import annotations

from enum import Enum, StrEnum


class StandardResultName(StrEnum):
    """The standard result name of a given data object."""

    field_outline = "field_outline"
    inplace_volumes = "inplace_volumes"
    structure_depth_surface = "structure_depth_surface"
    structure_time_surface = "structure_time_surface"
    structure_depth_isochore = "structure_depth_isochore"
    structure_depth_fault_lines = "structure_depth_fault_lines"
    structure_depth_fault_surface = "structure_depth_fault_surface"
    fluid_contact_surface = "fluid_contact_surface"
    fluid_contact_outline = "fluid_contact_outline"


class InplaceVolumes:
    """Enumerations relevant to inplace volumes tables."""

    class Fluid(str, Enum):
        """Fluid types used as values in the FLUID column."""

        oil = "oil"
        gas = "gas"
        water = "water"

    class TableIndexColumns(str, Enum):
        """The index columns for an inplace volumes table."""

        FLUID = "FLUID"
        ZONE = "ZONE"
        REGION = "REGION"
        FACIES = "FACIES"
        LICENSE = "LICENSE"

    class VolumetricColumns(str, Enum):
        """The value columns for an inplace volumes table."""

        BULK = "BULK"
        NET = "NET"
        PORV = "PORV"
        HCPV = "HCPV"
        STOIIP = "STOIIP"
        GIIP = "GIIP"
        ASSOCIATEDGAS = "ASSOCIATEDGAS"
        ASSOCIATEDOIL = "ASSOCIATEDOIL"

    @staticmethod
    def index_columns() -> list[str]:
        """Returns a list of the index columns."""
        return [k.value for k in InplaceVolumes.TableIndexColumns]

    @staticmethod
    def required_index_columns() -> list[str]:
        return [
            InplaceVolumes.TableIndexColumns.FLUID.value,
            InplaceVolumes.TableIndexColumns.ZONE.value,
            InplaceVolumes.TableIndexColumns.REGION.value,
        ]

    @staticmethod
    def value_columns() -> list[str]:
        """Returns a list of the value columns."""
        return [k.value for k in InplaceVolumes.VolumetricColumns]

    @staticmethod
    def required_value_columns() -> list[str]:
        """Returns a list of the value columns."""
        return [
            InplaceVolumes.VolumetricColumns.BULK.value,
            InplaceVolumes.VolumetricColumns.NET.value,
            InplaceVolumes.VolumetricColumns.PORV.value,
            InplaceVolumes.VolumetricColumns.HCPV.value,
        ]

    @staticmethod
    def required_columns() -> list[str]:
        """Returns a list of the columns required at export."""
        return (
            InplaceVolumes.required_index_columns()
            + InplaceVolumes.required_value_columns()
        )

    @staticmethod
    def table_columns() -> list[str]:
        """Returns a list of all table columns."""
        return InplaceVolumes.index_columns() + InplaceVolumes.value_columns()


class FaultLines:
    """Enumerations relevant to fault lines tables."""

    class TableIndexColumns(str, Enum):
        """The index columns for a fault lines table."""

        POLY_ID = "POLY_ID"
        NAME = "NAME"

    @staticmethod
    def index_columns() -> list[str]:
        """Returns a list of the index columns."""
        return [k.value for k in FaultLines.TableIndexColumns]



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from pydantic import BaseModel, Field, RootModel

from fmu.datamodels._schema_base import FmuSchemas, SchemaBase
from fmu.datamodels.standard_results.enums import InplaceVolumes
from fmu.datamodels.types import VersionStr

if TYPE_CHECKING:
    from typing import Any


class InplaceVolumesResultRow(BaseModel):
    """Represents the columns of a row in a static inplace volumes export.

    These fields are the current agreed upon standard result. Changes to the fields or
    their validation should cause the version defined in the standard result schema to
    increase the version number in a way that corresponds to the schema versioning
    specification (i.e. they are a patch, minor, or major change)."""

    FLUID: InplaceVolumes.Fluid
    """Index column. The kind of fluid this row represents. Typically GAS, OIL, or
    WATER. Required.
    """

    ZONE: str
    """Index column. The zone from which this volume is coming from. Required."""

    REGION: str
    """Index column. The region from which this volume is coming from. Required."""

    FACIES: str | None = Field(default=None)
    """Index column. The facies from which is volume is coming from. Optional."""

    LICENSE: str | None = Field(default=None)
    """Index column. The license under which these volumes related to. Optional."""

    BULK: float = Field(ge=0.0)
    """The bulk volume of the fluid-type given in ``FLUID``. Required."""

    NET: float = Field(ge=0.0)
    """The net volume of the fluid-type given in ``FLUID``. Required."""

    PORV: float = Field(ge=0.0)
    """The pore volume of the fluid-type given in ``FLUID``. Required."""

    HCPV: float | None = Field(default=None, ge=0.0)
    """The pore volume of the fluid-type given in ``FLUID``. Optional."""

    STOIIP: float | None = Field(default=None, ge=0.0)
    """The STOIIP volume of the fluid-type given in ``FLUID``. Optional."""

    GIIP: float | None = Field(default=None, ge=0.0)
    """The GIIP volume of the fluid-type given in ``FLUID``. Optional."""

    ASSOCIATEDGAS: float | None = Field(default=None, ge=0.0)
    """The associated gas volume of the fluid-type given in ``FLUID``. Optional."""

    ASSOCIATEDOIL: float | None = Field(default=None, ge=0.0)
    """The associated oil volume of the fluid-type given in ``FLUID``. Optional."""


class InplaceVolumesResult(RootModel):
    """Represents the resultant static inplace volumes parquet file, which is naturally
    a list of rows.

    Consumers who retrieve this parquet file must read it into a json-dictionary
    equivalent format to validate it against the schema."""

    root: list[InplaceVolumesResultRow]


class InplaceVolumesSchema(SchemaBase):
    """This class represents the schema that is used to validate the inplace volumes
    table being exported. This means that the version, schema filename, and schema
    location corresponds directly with the values and their validation constraints,
    documented above."""

    VERSION: VersionStr = "0.1.0"

    VERSION_CHANGELOG: str = """
    #### 0.1.0

    This is the initial schema version.
    """

    FILENAME: str = "inplace_volumes.json"
    """The filename this schema is written to."""

    PATH: Path = FmuSchemas.PATH / "file_formats" / VERSION / FILENAME
    """The local and URL path of this schema."""

    @classmethod
    def dump(cls) -> dict[str, Any]:
        return InplaceVolumesResult.model_json_schema(
            schema_generator=cls.default_generator()
        )



"""
Provides classes for managing and validating global configuration
settings in an application. These classes ensure essential settings
are defined and maintained consistently.
"""

from __future__ import annotations

import warnings
from typing import Any

from pydantic import (
    BaseModel,
    Field,
    RootModel,
    ValidationError,
    field_validator,
    model_validator,
)

from . import data, enums, fields


def validation_error_warning(err: ValidationError) -> None:
    """
    Emits a warning when a ValidationError is encountered in global configuration.
    """

    warnings.warn(
        f"""The global configuration has one or more errors that makes it
impossible to create valid metadata. The data will still be exported but no
metadata will be made. You are strongly encouraged to correct your
configuration. Invalid configuration may be disallowed in future versions.

Detailed information:
{str(err)}
""",
        stacklevel=2,
    )


class Ssdl(BaseModel):
    """
    Defines the configuration for the SSDL.
    """

    access_level: enums.Classification | None = Field(default=None)
    rep_include: bool | None = Field(default=None)


class Access(BaseModel, use_enum_values=True):
    """
    Manages access configurations, combining asset and SSDL information.
    """

    asset: fields.Asset
    ssdl: Ssdl | None = Field(default=None)
    classification: enums.Classification | None = Field(default=None)

    @model_validator(mode="after")
    def _validate_classification_ssdl_access_level(self) -> Access:
        if self.classification and self.ssdl and self.ssdl.access_level:
            warnings.warn(
                "The config contains both 'access.ssdl.access_level (deprecated) and "
                "access.classification. The value from access.classification will be "
                "used as the default classification. Remove 'access.ssdl.access_level' "
                "to silence this warning."
            )
        if not self.classification:
            if not (self.ssdl and self.ssdl.access_level):
                raise ValueError(
                    "The config doesn't contain any default security classification. "
                    "Please provide access.classification."
                )
            # classification mirrors ssdl.access_level if not present
            self.classification = self.ssdl.access_level

        return self


class StratigraphyElement(BaseModel):
    """
    Represents a single element in a stratigraphy configuration.
    """

    name: str
    stratigraphic: bool = Field(default=False)
    alias: list[str] | None = Field(default_factory=list)
    stratigraphic_alias: list[str] | None = Field(default=None)
    offset: float = Field(default=0.0, allow_inf_nan=False)
    top: data.Layer | None = Field(default=None)
    base: data.Layer | None = Field(default=None)

    @field_validator("alias", "stratigraphic_alias", mode="before")
    @classmethod
    def _prune_nones_and_adjust_input(cls, values: Any) -> Any:
        # For backwards compatibility, remove after a deprecation period
        if isinstance(values, list) and not all(values):
            warnings.warn(
                "The global config contains an empty list element in one of the "
                "'alias' fields in the 'stratigraphy' section. Please remove the empty "
                "element, and be aware that this will not be supported in the future.",
                FutureWarning,
            )
            return [v for v in values if v is not None]

        if isinstance(values, str):
            warnings.warn(
                "The global config contains string input for one of the 'alias' fields "
                "in the 'stratigraphy' section. Please convert to a list instead as "
                "this will not be supported in the future.",
                FutureWarning,
            )
            return [values]
        return values

    @field_validator("top", "base", mode="before")
    @classmethod
    def _set_name_attribute_if_string_input(
        cls, value: dict | str | None
    ) -> dict | None:
        if isinstance(value, str):
            return {"name": value}
        return value


class Stratigraphy(RootModel[dict[str, StratigraphyElement]]):
    """
    A collection of StratigraphyElement instances, accessible by keys.
    """

    def __iter__(self) -> Any:
        # Using ´Any´ as return type here as mypy is having issues
        # resolving the correct type
        return iter(self.root)

    def __getitem__(self, item: str) -> StratigraphyElement:
        return self.root[item]


class GlobalConfiguration(BaseModel):
    """
    Validates and manages the global configuration for the application.
    """

    access: Access
    masterdata: fields.Masterdata
    model: fields.Model
    stratigraphy: Stratigraphy | None = Field(default=None)



from __future__ import annotations

import os
from abc import ABC, abstractmethod
from collections.abc import Mapping
from pathlib import Path
from typing import (
    Any,
    Final,
    Literal,
    TypeVar,
)

from pydantic import BaseModel, ValidationError
from pydantic.json_schema import GenerateJsonSchema

from fmu.datamodels.types import VersionStr

T = TypeVar("T", dict, list, object)


class FmuSchemas:
    """These URLs can be constructed programmatically from radixconfig.yaml if need be:

        {cfg.components[].name}-{cfg.metadata.name}-{spec.environments[].name}

    As they are unlikely to change they are hardcoded here.
    """

    DEV_URL: Final[str] = "https://main-fmu-schemas-dev.radix.equinor.com"
    PROD_URL: Final[str] = "https://main-fmu-schemas-prod.radix.equinor.com"
    PATH: Final[Path] = Path("schemas")


class GenerateJsonSchemaBase(GenerateJsonSchema):
    """Implements a schema generator so that some additional fields may be
    added.

    This class also collects static methods used to transform the default OpenAPI
    schemas generated by Pydantic into schemas compatible with JSON Schema specs."""

    @staticmethod
    def remove_discriminator_mapping(data: T) -> T:
        """
        Removes entries with key ["discriminator"]["mapping"] from the schema. This
        adjustment is necessary because JSON Schema does not recognize this value
        while OpenAPI does.
        """

        if isinstance(data, dict):
            if "discriminator" in data and isinstance(data["discriminator"], dict):
                data["discriminator"].pop("mapping", None)

            for key, value in data.items():
                data[key] = GenerateJsonSchemaBase.remove_discriminator_mapping(value)

        elif isinstance(data, list):
            for index, element in enumerate(data):
                data[index] = GenerateJsonSchemaBase.remove_discriminator_mapping(
                    element
                )

        return data

    @staticmethod
    def remove_format_path(data: T) -> T:
        """
        Removes entries with key ["format"] = "path" from the schema. This
        adjustment is necessary because JSON Schema does not recognize the "format":
        "path", while OpenAPI does. This function is used in contexts where OpenAPI
        specifications are not applicable.
        """

        if isinstance(data, dict):
            return {
                k: GenerateJsonSchemaBase.remove_format_path(v)
                for k, v in data.items()
                if not (k == "format" and v == "path")
            }

        if isinstance(data, list):
            return [
                GenerateJsonSchemaBase.remove_format_path(element) for element in data
            ]

        return data

    def generate(
        self,
        schema: Mapping[str, Any],
        mode: Literal["validation", "serialization"] = "validation",
    ) -> dict[str, Any]:
        json_schema = super().generate(schema, mode=mode)

        json_schema = GenerateJsonSchemaBase.remove_discriminator_mapping(json_schema)
        json_schema = GenerateJsonSchemaBase.remove_format_path(json_schema)
        json_schema["$schema"] = self.schema_dialect

        return json_schema


class SchemaBase(ABC):
    VERSION: VersionStr
    """The current version of the schema."""

    VERSION_CHANGELOG: str
    """The changelog for all versions of schemas."""

    FILENAME: str
    """The filename, i.e. schema.json."""

    PATH: Path
    """The on-disk _and_ URL path following the domain, i.e:

        schemas/0.1.0/schema.json

    This path should _always_ have `FmuSchemas.PATH` as its first parent.
    This determines the on-disk and URL location of this schema file. A
    trivial example is:

        PATH: Path = FmuSchemas.PATH / VERSION / FILENAME

    """

    @classmethod
    def __init_subclass__(cls, **kwargs: dict[str, Any]) -> None:
        """This achieves Pydantic-like validation without being Pydantic.

        The reason for not using Pydantic here is that it does not play well with
        the class methods that are derived, and the dump method that must be
        implemented. It also doesn't like the default generator.
        """
        super().__init_subclass__(**kwargs)
        cls._validate_class_vars_set()
        cls._validate_version()
        cls._validate_version_changelog()
        cls._validate_path()

    @classmethod
    def _validate_class_vars_set(cls) -> None:
        for attr in ("VERSION", "VERSION_CHANGELOG", "FILENAME", "PATH"):
            if not hasattr(cls, attr):
                raise TypeError(f"Subclass {cls.__name__} must define '{attr}'")

    @classmethod
    def _validate_path(cls) -> None:
        if not cls.PATH.parts[0].startswith(str(FmuSchemas.PATH)):
            raise ValueError(
                f"PATH must start with `FmuSchemas.PATH`: {FmuSchemas.PATH}. "
                f"Got {cls.PATH}"
            )

    @classmethod
    def _validate_version(cls) -> None:
        try:

            class PydanticVersionValidator(BaseModel):
                """VersionStr uses a Pydantic 'Field'. Use Pydantic to validate it."""

                version: VersionStr

            PydanticVersionValidator(version=cls.VERSION)
        except ValidationError as e:
            raise TypeError(f"Invalid VERSION format for '{cls.__name__}': {e}") from e

    @classmethod
    def _validate_version_changelog(cls) -> None:
        """Ensures that VERSION has a doc string that looks like a changelog.

        Checks that the current version has an entry as well."""
        if f"### {cls.VERSION}" not in cls.VERSION_CHANGELOG:
            raise ValueError(
                f"No changelog entry exists for '{cls.__name__}' version {cls.VERSION}."
                f" Expected a header like '### {cls.VERSION}' in 'VERSION_CHANGELOG' "
                f"with changes for version {cls.VERSION} listed beneath it."
            )

    @classmethod
    def dev_url(cls) -> str:
        """Returns the url to the schema on the Radix dev environment."""
        return f"{FmuSchemas.DEV_URL}/{cls.PATH}"

    @classmethod
    def prod_url(cls) -> str:
        """Returns the url to the schema on the Radix prod environment."""
        return f"{FmuSchemas.PROD_URL}/{cls.PATH}"

    @classmethod
    def url(cls) -> str:
        """Returns the URL the schema resides at.

        Schema URLs are composed from class variables set by children derived from this
        base class. If we're in a development environment, or Komodo bleeding, return
        the dev schema URL."""
        if os.environ.get("DEV_SCHEMA", None) or "bleeding" in os.environ.get(
            "KOMODO_RELEASE", os.environ.get("KOMODO_RELEASE_BACKUP", "")
        ):
            return cls.dev_url()
        return cls.prod_url()

    @classmethod
    def default_generator(cls) -> type[GenerateJsonSchema]:
        """Provides a default schema generator that should be adequate for most simple
        schemas.

        When more customization is required a separate schema generator may be
        warranted. See the 'FmuResults' model for how this can be done."""

        class DefaultGenerateJsonSchema(GenerateJsonSchemaBase):
            """Implements a schema generator so that some additional fields may be
            added."""

            def generate(
                self,
                schema: Mapping[str, Any],
                mode: Literal["validation", "serialization"] = "validation",
            ) -> dict[str, Any]:
                json_schema = super().generate(schema, mode=mode)

                json_schema["$id"] = cls.url()
                json_schema["version"] = cls.VERSION

                return json_schema

        return DefaultGenerateJsonSchema

    @classmethod
    @abstractmethod
    def dump(cls) -> dict[str, Any]:
        """
        Dumps the export root model to JSON format for schema validation and
        usage in FMU data structures.

        To update the schema:
            1. Run the following CLI command to dump the updated schema:
                `./tools/update-schemas --diff`.
            2. Check the diff for changes. Adding fields usually indicates non-breaking
                changes and is generally safe. However, if fields are removed, it could
                indicate breaking changes that may affect dependent systems. Perform a
                quality control (QC) check to ensure these changes do not break existing
                implementations.
                If changes are satisfactory and do not introduce issues, commit
                them to maintain schema consistency.
        """
        raise NotImplementedError



from __future__ import annotations

import json
from typing import Any

import pytest
from pytest import MonkeyPatch

from fmu.datamodels import schemas
from fmu.datamodels._schema_base import FmuSchemas, SchemaBase


def _contains_discriminator_mapping(schema: Any) -> bool:
    """Recursively checks ["discriminator"]["mapping"] in the schema."""

    if isinstance(schema, dict):
        if (
            "discriminator" in schema and isinstance(schema["discriminator"], dict)
        ) and "mapping" in schema["discriminator"]:
            return True
        for value in schema.values():
            if _contains_discriminator_mapping(value):
                return True
    elif isinstance(schema, list):
        for item in schema:
            if _contains_discriminator_mapping(item):
                return True
    return False


@pytest.mark.parametrize("schema", schemas)
def test_schemas_uptodate(schema: SchemaBase) -> None:
    """
    Test to verify if the local schemas are up to date with the schema
    generated by pydantic's `dump` method. It compares the content of
    the local schema with the output of `dump()`.

    To get more feedback or generate new schemas run:

        ./tools/update-schemas --diff

    If you are generating a production release try running:

        ./tools/update-schemas --diff --prod
    """

    with open(schema.PATH) as f:
        assert json.load(f) == schema.dump()


@pytest.mark.parametrize("schema", schemas)
def test_schema_url_changes_with_env_var(
    schema: SchemaBase, monkeypatch: MonkeyPatch
) -> None:
    monkeypatch.setenv("DEV_SCHEMA", "")
    json = schema.dump()
    assert schema.url().startswith(FmuSchemas.PROD_URL)
    assert json["$id"].startswith(FmuSchemas.PROD_URL)
    assert json["$schema"] == "https://json-schema.org/draft/2020-12/schema"

    monkeypatch.setenv("DEV_SCHEMA", "1")
    json = schema.dump()
    assert schema.url().startswith(FmuSchemas.DEV_URL)
    assert json["$id"].startswith(FmuSchemas.DEV_URL)
    assert json["$schema"] == "https://json-schema.org/draft/2020-12/schema"


@pytest.mark.parametrize("schema", schemas)
@pytest.mark.parametrize("env_var", ["KOMODO_RELEASE", "KOMODO_RELEASE_BACKUP"])
@pytest.mark.parametrize(
    "env_var_value",
    [
        "bleeding",
        "kenv_bleeding",
        "bleeding-20250306-0207-py311-rhel8",
    ],
)
def test_schema_url_changes_with_komodo_bleeding_env_var(
    schema: SchemaBase, env_var: str, env_var_value: str, monkeypatch: MonkeyPatch
) -> None:
    """Tests that the schema uses the development url when KOMODO_RELEASE or
    KOMODO_RELEASE_BACKUP is set to a bleeding release.

    DEV_SCHEMA is set to 1 whenever the tests are run, so it is unset here."""
    monkeypatch.delenv("DEV_SCHEMA", raising=False)
    json = schema.dump()
    assert schema.url().startswith(FmuSchemas.PROD_URL)
    assert json["$id"].startswith(FmuSchemas.PROD_URL)
    assert json["$schema"] == "https://json-schema.org/draft/2020-12/schema"

    # needed to prevent tests failing if inside an actual komodo environment
    monkeypatch.delenv("KOMODO_RELEASE", raising=False)

    monkeypatch.setenv(env_var, env_var_value)
    json = schema.dump()
    assert schema.url().startswith(FmuSchemas.DEV_URL)
    assert json["$id"].startswith(FmuSchemas.DEV_URL)
    assert json["$schema"] == "https://json-schema.org/draft/2020-12/schema"


@pytest.mark.parametrize("schema", schemas)
def test_no_discriminator_mappings_leftover_in_schema(schema: SchemaBase) -> None:
    """Sumo's AJV validator doesn't like discriminator mappings leftover in the
    schema."""
    with open(schema.PATH) as f:
        schema = json.load(f)
    assert _contains_discriminator_mapping(schema) is False



from __future__ import annotations

from enum import Enum, IntEnum, StrEnum


class Classification(StrEnum):
    """The security classification for a given data object."""

    asset = "asset"
    internal = "internal"
    restricted = "restricted"


class AxisOrientation(IntEnum):
    """The axis orientation for a given data object."""

    normal = 1
    flipped = -1


class Content(StrEnum):
    """The content type of a given data object."""

    depth = "depth"
    facies_thickness = "facies_thickness"
    fault_surface = "fault_surface"
    fault_lines = "fault_lines"
    fault_properties = "fault_properties"
    field_outline = "field_outline"
    field_region = "field_region"
    fluid_contact = "fluid_contact"
    khproduct = "khproduct"
    lift_curves = "lift_curves"
    named_area = "named_area"
    parameters = "parameters"
    pinchout = "pinchout"
    property = "property"
    pvt = "pvt"
    regions = "regions"
    relperm = "relperm"
    rft = "rft"
    seismic = "seismic"
    simulationtimeseries = "simulationtimeseries"
    subcrop = "subcrop"
    thickness = "thickness"
    time = "time"
    timeseries = "timeseries"
    transmissibilities = "transmissibilities"
    velocity = "velocity"
    volumes = "volumes"
    wellpicks = "wellpicks"

    @classmethod
    def _missing_(cls: type[Content], value: object) -> None:
        raise ValueError(
            f"Invalid 'content' {value=}. Valid entries are {[m.value for m in cls]}"
        )


class ErtSimulationMode(str, Enum):
    """The simulation mode ert was run in. These definitions come from
    `ert.mode_definitions`."""

    ensemble_experiment = "ensemble_experiment"
    ensemble_information_filter = "ensemble_information_filter"
    ensemble_smoother = "ensemble_smoother"
    es_mda = "es_mda"
    evaluate_ensemble = "evaluate_ensemble"
    manual_update = "manual_update"
    test_run = "test_run"
    workflow = "workflow"


class MetadataClass(StrEnum):
    """Base class for objects by FMU convention or standards."""


class ObjectMetadataClass(MetadataClass):
    """The class of a data object (typically originating from an RMS model)."""

    surface = "surface"
    table = "table"
    cpgrid = "cpgrid"
    cpgrid_property = "cpgrid_property"
    polygons = "polygons"
    cube = "cube"
    well = "well"
    points = "points"
    dictionary = "dictionary"


class FMUResultsMetadataClass(MetadataClass):
    """The class of an FMU results object."""

    case = "case"
    realization = "realization"
    iteration = "iteration"
    ensemble = "ensemble"


class Layout(StrEnum):
    """The layout of a given data object."""

    regular = "regular"
    unset = "unset"
    cornerpoint = "cornerpoint"
    table = "table"
    dictionary = "dictionary"
    faultroom_triangulated = "faultroom_triangulated"
    triangulated = "triangulated"


class FMUContext(str, Enum):
    """The context in which FMU was being run when data were generated."""

    case = "case"
    iteration = "iteration"
    ensemble = "ensemble"
    realization = "realization"


class VerticalDomain(StrEnum):
    depth = "depth"
    time = "time"


class DomainReference(StrEnum):
    msl = "msl"
    sb = "sb"
    rkb = "rkb"


class TrackLogEventType(StrEnum):
    """The type of event being logged"""

    created = "created"
    updated = "updated"
    merged = "merged"


class FluidContactType(StrEnum):
    """The type of fluid contact."""

    fgl = "fgl"
    """Free gas level."""

    fwl = "fwl"
    """Free water level."""

    goc = "goc"
    """Gas-oil contact."""

    gwc = "gwc"
    """Gas-water contact."""

    owc = "owc"
    """Oil-water contact."""


class FileFormat(StrEnum):
    """The format of a given data object."""

    parquet = "parquet"
    json = "json"
    csv = "csv"
    csv_xtgeo = "csv|xtgeo"
    irap_ascii = "irap_ascii"
    irap_binary = "irap_binary"
    roff = "roff"
    segy = "segy"
    openvds = "openvds"
    tsurf = "tsurf"
